from __future__ import absolute_import
from datetime import datetime
from flask.ext.sqlalchemy import orm

from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property


from extensions import db
from .utils import slugify, plural_underscored


class BaseMixin(object):
    """ Base mixin
    """
    __table_args__ = {'extend_existing': True,
                      'mysql_charset': 'utf8',
                      'mysql_engine': 'InnoDB'}

    @declared_attr
    def __tablename__(cls):
        """ We want our app to be more English for pluralization cases
        """
        return plural_underscored(cls.__name__)

    def save(self, commit=True):
        db.session.add(self)
        commit and db.session.commit()
        return self

    def delete(self, commit=True):
        db.session.delete(self)
        commit and db.session.commit()

    def __repr__(self):
        return "<%s:%s>" % (self.__class__.__name__,
                            self.id)


class CRUDMixin(BaseMixin):
    """ Basic CRUD mixin
    """

    @declared_attr
    def id(cls):
        return db.Column(db.Integer, primary_key=True)

    @declared_attr
    def created_at(cls):
        return db.Column(db.DateTime, default=datetime.utcnow)

    @classmethod
    def get(cls, id):
        return cls.query.get(id)

    @classmethod
    def create(cls, commit=True, **kwargs):
        return cls(**kwargs).save(commit)

    def update(self, commit=True, **kwargs):
        return self._setattrs(**kwargs).save(commit)

    def as_dict(self, api_fields=None, exclude=None):
        """ method for building dictionary for model value-properties filled
            with data from mapped storage backend
        """
        exclude = set(exclude or [])
        api_fields = api_fields or getattr(self, 'api_fields', [])

        column_properties = [p.key for p in self.__mapper__.iterate_properties
                                if isinstance(p, orm.ColumnProperty)]

        exportable_fields = set(column_properties.extend(api_fields))

        # convert undescored fields:
        fields = set(map(lambda f: f.strip('_'), exportable_fields))

        exportable_fields = fields - exclude

        results = dict()

        for field in exportable_fields:
            attr = getattr(self, field)
            # process callable objects
            value = hasattr(attr, '__call__') and attr() or attr
            results[field] = value

        return results

    def _setattrs(self, **kwargs):
        for k, v in kwargs.iteritems():
            k.startswith('_') and raise_value('Underscored values are not'
                                              ' allowed')
            setattr(self, k, v)

        return self


class SlugMixin(CRUDMixin):
    """Basic mixin for models with slug and name
    """
    @declared_attr
    def title(cls):
        return db.Column(db.Unicode(512), nullable=False)

    @declared_attr
    def _slug(cls):
        return db.Column(db.String(128), nullable=False, unique=True)

    @hybrid_property
    def slug(self):
        return self._slug

    @slug.setter
    def slug(self, title):
        self._slug = slugify(title)

    def save(self, commit=True):
        self.slug = self.title
        return super(SlugMixin, self).save(commit)

    @classmethod
    def get_by_slug(cls, slug):
        return cls.query.filter_by(slug=slug).first()


def raise_value(text):
    """ Simple helper to raise exception in boolean expressions"""
    raise ValueError(text)
